

set -ex



pip check
pytest -v test/unit  --ignore=test/unit/test_configmanager.py --ignore=test/unit/test_connection.py --ignore=test/unit/test_encryption_util.py --ignore=test/unit/test_error_arrow_stream.py --ignore=test/unit/test_result_batch.py --ignore=test/unit/test_s3_util.py --ignore=test/unit/test_retry_network.py  --deselect=test/unit/test_cache.py::TestSFDictFileCache::test_read_only --deselect=test/unit/test_easy_logging.py::test_config_file_inaccessible_path --deselect=test/unit/test_put_get.py::test_put_error
exit 0
